"""Inverting pair mappings helper tool."""

import os
import pathlib

from typing import Union  # py39 does not handle |

__version__ = '2025.7.19'
__version_info__ = tuple(
    e if '-' not in e else e.split('-')[0] for part in __version__.split('+') for e in part.split('.') if e != 'parent'
)
__all__: list[str] = [
    'APP_ALIAS',
    'APP_NAME',
    'DEBUG',
    'ENCODING',
    'ENC_ERRS',
    'GENERATOR_CAVEAT_KEY',
    'GENERATOR_CAVEAT_VALUE',
    'JSON_FORMAT',
    'LOG_AMBIGUITY',
    'LOG_MISSING',
    'PathLike',
    'VERSION',
    'VERSION_INFO',
    'YAML_FORMAT',
]

APP_ALIAS = str(pathlib.Path(__file__).parent.name)
APP_ENV = APP_ALIAS.upper()
APP_NAME = locals()['__doc__']

DEBUG = bool(os.getenv(f'{APP_ENV}_DEBUG', ''))

ENCODING = 'utf-8'
ENC_ERRS = 'ignore'

GENERATOR_CAVEAT_KEY = 'Please do not edit manually!'
GENERATOR_CAVEAT_VALUE = 'Cf. documentation'

LOG_AMBIGUITY = 'Error: source has ambiguous values.'
LOG_MISSING = 'Error: source file does not exist.'

JSON_FORMAT = 'json'
YAML_FORMAT = 'yaml'

PathLike = Union[pathlib.Path, str]
VERSION = __version__
VERSION_INFO = __version_info__



import pathlib
import tempfile

import pytest  # type: ignore

import inverso.cli as cli
from inverso import ENCODING, ENC_ERRS, VERSION

FIXTURE = pathlib.Path('test/fixtures/')
USAGE_STARTS_WITH = 'usage: inverso [-h] [--source FILE] [--target FILE] [--debug] [--auto-serial]'

REPRODUCIBLE_PREVIEW_LINES = """Preview of inversion request (dry-run):
  source_pos: None
  target_pos: None
  source: test/fixtures/some-object.yml
  debug: False
  auto_serial: False
  marker_token:
  marker_is_value: False
  generator_caveat: False
  preview: True
  quiet: True
  version: False
  source_format: yaml
  target_format: yaml
  auto_serial_start: 0
  auto_serial_step: 1
""".split()

DYNAMIC_PREVIEW_LINE_STARTS_WITH = '  target:'
DYNAMIC_PREVIEW_LINE_ENDS_WITH = '/away-you-object-thing.yaml'


def test_main_ok_empty(capsys):
    cli.main([])
    out, err = capsys.readouterr()
    assert not err
    assert 'show this help message and exit' in out
    assert 'JSON' in out


def test_main_ok_smvp(capsys):
    with pytest.raises(SystemExit, match='0'):
        cli.main(['-h'])
    out, err = capsys.readouterr()
    assert not err
    assert 'show this help message and exit' in out
    assert 'JSON' in out


def test_parse_request_version(capsys):
    options = cli.parse_request(['--version'])
    assert options == 0
    out, err = capsys.readouterr()
    assert not err
    assert out.rstrip() == VERSION


def test_main_nok_no_files(capsys):
    with pytest.raises(SystemExit, match='2'):
        cli.main(['-d'])
    out, err = capsys.readouterr()
    assert err.startswith(USAGE_STARTS_WITH)
    assert err.rstrip().endswith(
        'inverso: error: source path must be given'
        ' - either as first positional argument or as value to the --source option'
    )
    assert not out


def test_main_nok_debug_and_quiet(capsys):
    with pytest.raises(SystemExit, match='2'):
        cli.main(['--debug', '--quiet'])
    out, err = capsys.readouterr()
    assert err.startswith(USAGE_STARTS_WITH)
    assert err.rstrip().endswith('inverso: error: Cannot be quiet and debug - pick one')
    assert not out


def test_main_nok_no_target_file(capsys):
    with pytest.raises(SystemExit, match='2'):
        cli.main([str(FIXTURE / 'empty.json')])
    out, err = capsys.readouterr()
    assert err.startswith(USAGE_STARTS_WITH)
    assert err.rstrip().endswith(
        'error: target path must be given - either as second positional argument or as value to the --target option'
    )
    assert not out


def test_main_nok_no_source_file(capsys):
    with pytest.raises(SystemExit, match='2'):
        cli.main(['--target', str(FIXTURE / 'not-present.json')])
    out, err = capsys.readouterr()
    assert err.startswith(USAGE_STARTS_WITH)
    assert err.rstrip().endswith(
        'error: source path must be given - either as first positional argument or as value to the --source option'
    )
    assert not out


def test_main_nok_source_file_not_present(capsys):
    source = FIXTURE / 'not-present.json'
    with pytest.raises(SystemExit, match='2'):
        cli.main(['--source', str(source)])
    out, err = capsys.readouterr()
    assert err.startswith(USAGE_STARTS_WITH)
    assert err.rstrip().endswith(f'error: requested source ({source}) is not a file')
    assert not out


def test_main_nok_source_file_has_unknown_suffix(capsys):
    target = FIXTURE / 'not-present.json'
    with pytest.raises(SystemExit, match='2'):
        cli.main([str(FIXTURE / 'empty.jason'), str(target)])
    out, err = capsys.readouterr()
    # assert err.startswith(USAGE_STARTS_WITH)
    assert err.rstrip().endswith(
        'error: requested source suffix (.jason) is not in known suffixes (.json, .yaml, .yml)'
    )
    assert not out


def test_main_nok_doubled_source_files(capsys):
    source = FIXTURE / 'not-present.json'
    with pytest.raises(SystemExit, match='2'):
        cli.main(['--source', str(source), str(source)])
    out, err = capsys.readouterr()
    assert err.startswith(USAGE_STARTS_WITH)
    assert err.rstrip().endswith(
        'error: source path given both as first positional argument and as value to the --source option - pick one'
    )
    assert not out


def test_main_nok_target_is_source_file(capsys):
    target = FIXTURE / 'empty.json'
    with pytest.raises(SystemExit, match='2'):
        cli.main([str(FIXTURE / 'empty.json'), str(target)])
    out, err = capsys.readouterr()
    assert err.startswith(USAGE_STARTS_WITH)
    assert err.rstrip().endswith(
        'error: target path is identical with source path - cowardly giving up on the attempted in-place inversion'
    )
    assert not out


def test_main_nok_target_file_has_unknown_suffix(capsys):
    target = FIXTURE / 'empty-array.jason'
    with pytest.raises(SystemExit, match='2'):
        cli.main([str(FIXTURE / 'empty.json'), str(target)])
    out, err = capsys.readouterr()
    assert err.startswith(USAGE_STARTS_WITH)
    assert err.rstrip().endswith(
        'error: requested target suffix (.jason) is not in known suffixes (.json, .yaml, .yml)'
    )
    assert not out


def test_main_nok_doubled_target_files(capsys):
    source = FIXTURE / 'empty.json'
    target = FIXTURE / 'not-present.json'
    with pytest.raises(SystemExit, match='2'):
        cli.main([str(source), '--target', str(target), str(target)])
    out, err = capsys.readouterr()
    assert err.startswith(USAGE_STARTS_WITH)
    assert err.rstrip().endswith(
        'error: target path given both as second positional argument and as value to the --target option - pick one'
    )
    assert not out


def test_main_nok_nonjective_source(capsys):
    source = FIXTURE / 'nonjective.json'
    target = FIXTURE / 'not-present.json'
    # with pytest.raises(SystemExit, match='1'):
    cli.main([str(source), '--target', str(target)])
    out, err = capsys.readouterr()
    # assert err.startswith(USAGE_STARTS_WITH)
    assert err.rstrip().endswith("Error: source has ambiguous values.\n- value='c' occurs for keys=['a', 'b']")
    assert not out


def test_main_ok_some_json_object_to_yaml(capsys):
    source = FIXTURE / 'some-object.json'
    with tempfile.TemporaryDirectory() as temp_dir:
        target = pathlib.Path(temp_dir) / 'away-you-object-thing.yml'
        cli.main(['-d', str(source), str(target)])
        assert target.is_file()
        with open(target, 'rt', encoding=ENCODING, errors=ENC_ERRS) as handle:
            result = handle.read()
        with open(FIXTURE / 'some-object.yml', 'rt', encoding=ENCODING, errors=ENC_ERRS) as handle:
            expected = handle.read()
        assert result == expected
    out, err = capsys.readouterr()
    assert not err
    assert out.startswith('Debug mode requested.')
    assert 'Requested inversion from json to yaml' in out
    assert 'json_load(path=test/fixtures/some-object.json, debug=True)' in out
    assert 'away-you-object-thing.yml' in out


def test_main_ok_auto_serialize_json_object_to_json_without_generator_caveat(capsys):
    marker = '4321'
    source = FIXTURE / f'auto-serial-marker-{marker}.json'
    target_ref = FIXTURE / f'auto-serial-marker-{marker}-inverted-without-generator-caveat.json'
    with tempfile.TemporaryDirectory() as temp_dir:
        target = pathlib.Path(temp_dir) / 'away-you-object-thing.json'
        cli.main(['-am', marker, str(source), str(target)])
        assert target.is_file()
        with open(target, 'rt', encoding=ENCODING, errors=ENC_ERRS) as handle:
            result = handle.read()
        with open(target_ref, 'rt', encoding=ENCODING, errors=ENC_ERRS) as handle:
            expected = handle.read()
        assert result == expected
    out, err = capsys.readouterr()
    assert not err
    assert not out


def test_main_ok_auto_serialize_json_object_to_json_with_generator_caveat(capsys):
    marker = '4321'
    source = FIXTURE / f'auto-serial-marker-{marker}.json'
    target_ref = FIXTURE / f'auto-serial-marker-{marker}-inverted-with-generator-caveat.json'
    with tempfile.TemporaryDirectory() as temp_dir:
        target = pathlib.Path(temp_dir) / 'away-you-object-thing.json'
        cli.main(['-agq', '-m', marker, str(source), str(target)])
        assert target.is_file()
        with open(target, 'rt', encoding=ENCODING, errors=ENC_ERRS) as handle:
            result = handle.read()
        with open(target_ref, 'rt', encoding=ENCODING, errors=ENC_ERRS) as handle:
            expected = handle.read()
        assert result == expected
    out, err = capsys.readouterr()
    assert not err
    assert not out


def test_main_ok_auto_serialize_json_object_no_marker_to_json_without_generator_caveat(capsys):
    source = FIXTURE / 'auto-serial-no-marker.json'
    target_ref = FIXTURE / 'auto-serial-no-marker-inverted-without-generator-caveat.json'
    with tempfile.TemporaryDirectory() as temp_dir:
        target = pathlib.Path(temp_dir) / 'away-you-object-thing.json'
        cli.main(['-a', str(source), str(target)])
        assert target.is_file()
        with open(target, 'rt', encoding=ENCODING, errors=ENC_ERRS) as handle:
            result = handle.read()
        with open(target_ref, 'rt', encoding=ENCODING, errors=ENC_ERRS) as handle:
            expected = handle.read()
        assert result == expected
    out, err = capsys.readouterr()
    assert not err
    assert not out


def test_main_ok_auto_serialize_json_object_no_marker_to_json_with_generator_caveat(capsys):
    source = FIXTURE / 'auto-serial-no-marker.json'
    target_ref = FIXTURE / 'auto-serial-no-marker-inverted-with-generator-caveat.json'
    with tempfile.TemporaryDirectory() as temp_dir:
        target = pathlib.Path(temp_dir) / 'away-you-object-thing.json'
        cli.main(['-agq', str(source), str(target)])
        assert target.is_file()
        with open(target, 'rt', encoding=ENCODING, errors=ENC_ERRS) as handle:
            result = handle.read()
        with open(target_ref, 'rt', encoding=ENCODING, errors=ENC_ERRS) as handle:
            expected = handle.read()
        assert result == expected
    out, err = capsys.readouterr()
    assert not err
    assert not out


def test_main_ok_some_yaml_object_to_json(capsys):
    source = FIXTURE / 'some-object.yml'
    with tempfile.TemporaryDirectory() as temp_dir:
        target = pathlib.Path(temp_dir) / 'away-you-object-thing.json'
        cli.main(['-d', str(source), str(target)])
        assert target.is_file()
    out, err = capsys.readouterr()
    assert not err
    assert out.startswith('Debug mode requested.')
    assert 'Requested inversion from yaml to json' in out
    assert 'yaml_load(path=test/fixtures/some-object.yml, debug=True)' in out
    assert 'away-you-object-thing.json' in out


def test_main_ok_some_json_object_to_json(capsys):
    source = FIXTURE / 'some-object.json'
    with tempfile.TemporaryDirectory() as temp_dir:
        target = pathlib.Path(temp_dir) / 'away-you-object-thing.JSON'
        cli.main(['-d', str(source), str(target)])
        assert target.is_file()
    out, err = capsys.readouterr()
    assert not err
    assert out.startswith('Debug mode requested.')
    assert 'Requested inversion from json to json' in out
    assert 'json_load(path=test/fixtures/some-object.json, debug=True)' in out
    assert 'away-you-object-thing.json' in out.lower()


def test_main_ok_some_yaml_object_to_yaml(capsys):
    source = FIXTURE / 'some-object.yml'
    with tempfile.TemporaryDirectory() as temp_dir:
        target = pathlib.Path(temp_dir) / 'away-you-object-thing.yaml'
        cli.main(['-d', str(source), str(target)])
        assert target.is_file()
    out, err = capsys.readouterr()
    assert not err
    assert out.startswith('Debug mode requested.')
    assert 'Requested inversion from yaml to yaml' in out
    assert 'yaml_load(path=test/fixtures/some-object.yml, debug=True)' in out
    assert 'away-you-object-thing.yaml' in out


def test_main_ok_target_from_option(capsys):
    source = FIXTURE / 'some-object.yml'
    with tempfile.TemporaryDirectory() as temp_dir:
        target = pathlib.Path(temp_dir) / 'away-you-object-thing.yaml'
        cli.main(['-d', '-s', str(source), '-t', str(target)])
        assert target.is_file()
    out, err = capsys.readouterr()
    assert not err
    assert out.startswith('Debug mode requested.')
    assert 'Requested inversion from yaml to yaml' in out
    assert 'yaml_load(path=test/fixtures/some-object.yml, debug=True)' in out
    assert 'away-you-object-thing.yaml' in out


def test_main_ok_quiet(capsys):
    source = FIXTURE / 'some-object.yml'
    with tempfile.TemporaryDirectory() as temp_dir:
        target = pathlib.Path(temp_dir) / 'away-you-object-thing.yaml'
        cli.main(['-q', '-s', str(source), '-t', str(target)])
        assert target.is_file()
    out, err = capsys.readouterr()
    assert not err
    assert not out


def test_main_ok_preview(capsys):
    source = FIXTURE / 'some-object.yml'
    with tempfile.TemporaryDirectory() as temp_dir:
        target = pathlib.Path(temp_dir) / 'away-you-object-thing.yaml'
        cli.main(['-qp', '-s', str(source), '-t', str(target)])
        assert not target.is_file()
    out, err = capsys.readouterr()
    assert not err
    out_lines = out.split()
    for out_line in out_lines:
        if out_line.startswith(DYNAMIC_PREVIEW_LINE_STARTS_WITH):
            assert out_line.endswith(DYNAMIC_PREVIEW_LINE_ENDS_WITH)
        else:
            out_line in REPRODUCIBLE_PREVIEW_LINES



"""Inverting pair mappings API for opinionated transforms."""

import argparse
import sys

from muuntuu.implementation import json_dump, json_load, yaml_dump, yaml_load  # type: ignore
from inverso.implementation import nonjective, process
from inverso import GENERATOR_CAVEAT_KEY, GENERATOR_CAVEAT_VALUE, JSON_FORMAT, LOG_AMBIGUITY


def invert(request: argparse.Namespace) -> int:
    """Invert between and within the known formats."""
    _ = request.debug and print('Debug mode requested.')

    load_source = json_load if request.source_format == JSON_FORMAT else yaml_load
    dump_source = json_dump if request.source_format == JSON_FORMAT else yaml_dump
    dump_target = json_dump if request.target_format == JSON_FORMAT else yaml_dump

    _ = request.debug and print(f'Requested inversion from {request.source_format} to {request.target_format}')
    auto_serial = request.auto_serial_start
    cleansed: dict[str, str] = {}
    for k, v in load_source(request.source, request.debug).items():
        key, value, seen = process(k, v, auto_serial, request.marker_token, request.marker_is_value)
        cleansed[key] = value
        if auto_serial:
            if not request.marker_token or not seen:
                auto_serial += request.auto_serial_step

    dump_source(cleansed, request.source, request.debug)

    if findings := nonjective(cleansed):
        print(LOG_AMBIGUITY, file=sys.stderr)
        for finding in findings:
            print(f'- {finding}', file=sys.stderr)
        return 1

    inverted = {v: k for k, v in cleansed.items()}
    ordered = {GENERATOR_CAVEAT_KEY: GENERATOR_CAVEAT_VALUE} if request.generator_caveat else {}
    for k in sorted(inverted):
        ordered[k] = inverted[k]

    dump_target(ordered, request.target, request.debug)

    return 0



"""Format specific load and dump implementations for the inverting pair mappings helper tool."""


def process(key: str, value: str, serial: int, marker: str, is_value: bool) -> tuple[str, str, bool]:
    """Process the key value pair and signal any marker as either key or value.

    If serial is 1 or greater, enter auto-serial mode.
    When in auto-serial mode and if marker is not empty, then inspect exemptions.
    If marker is_value, assume the marker is in the value slot of the incoming pair.
    """
    if serial:  # Use auto-serial on keys
        if marker:  # Signal any marker value to be kept as is and pause the auto increment
            if any([not is_value and key == marker, is_value and value == marker]):
                return key, value, True
        return str(serial), value, False
    return key, value, False


def nonjective(data: dict[str, str]) -> list[str]:
    """Assess the uniqueness of values, so we can invert without ambiguity."""
    if sorted(set(data.values())) == sorted(data.values()):
        return []

    pairs = {}
    for k, v in data.items():
        if v not in pairs:
            pairs[v] = [k]
        else:
            pairs[v].append(k)

    findings = []
    for value, keys in pairs.items():
        findings.append(f'{value=} occurs for {keys=}')

    return findings



"""Command line interface for inverting pair mappings helper tool."""

import argparse
import pathlib
import sys

from typing import Union  # py39 does not handle |

from inverso.api import invert
from inverso import (
    APP_ALIAS,
    APP_ENV,
    APP_NAME,
    DEBUG,
    JSON_FORMAT,
    VERSION,
    YAML_FORMAT,
)

JSON_SUFFIXES = set(['.json'])
YAML_SUFFIXES = set(['.yaml', '.yml'])
KNOWN_SUFFIXES = JSON_SUFFIXES | YAML_SUFFIXES


def parser_configured() -> argparse.ArgumentParser:
    """Return the configured argument parser."""
    parser = argparse.ArgumentParser(
        prog=APP_ALIAS,
        description=APP_NAME,
        formatter_class=argparse.RawTextHelpFormatter,
    )
    parser.add_argument(
        'source_pos',
        nargs='?',
        metavar='SOURCE_FILE',
        help='JSON or YAML source as positional argument',
    )
    parser.add_argument(
        'target_pos',
        nargs='?',
        metavar='TARGET_FILE',
        help='JSON or YAML target as positional argument',
    )
    parser.add_argument(
        '--source',
        '-s',
        dest='source',
        metavar='FILE',
        help='JSON or YAML source',
    )
    parser.add_argument(
        '--target',
        '-t',
        dest='target',
        metavar='FILE',
        help='JSON or YAML target',
    )
    parser.add_argument(
        '--debug',
        '-d',
        dest='debug',
        default=None,  # not False, so we can distinguish environment variable overwrite
        action='store_true',
        help=f'work in debug mode (default: False), overwrites any environment variable {APP_ENV}_DEBUG value',
    )
    parser.add_argument(
        '--auto-serial',
        '-a',
        dest='auto_serial',
        default=False,
        action='store_true',
        help='auto-serial mode, rewrite incoming keys as 1-based auto-serial (default: False)',
    )
    parser.add_argument(
        '--marker-token',
        '-m',
        dest='marker_token',
        metavar='TOKEN',
        default='',
        help='if in auto-serial mode, marker token to be exempted (default: False)',
    )
    parser.add_argument(
        '--is-value',
        dest='marker_is_value',
        default=False,
        action='store_true',
        help='if marker token, then expect it as value insteadf of as key (default: False)',
    )
    parser.add_argument(
        '--generator-caveat',
        '-g',
        dest='generator_caveat',
        default=False,
        action='store_true',
        help='add a generator caveat as first pair to generated inverted map (default: False)',
    )
    parser.add_argument(
        '--preview',
        '-p',
        dest='preview',
        default=False,
        action='store_true',
        help='preview only (dry-run) reporting on what would be done (default: False)',
    )
    parser.add_argument(
        '--quiet',
        '-q',
        dest='quiet',
        default=False,
        action='store_true',
        help='work in quiet mode (default: False)',
    )
    parser.add_argument(
        '--version',
        '-V',
        dest='version',
        default=False,
        action='store_true',
        help='display version and exit',
    )

    return parser


def validate_source(options: argparse.Namespace, engine: argparse.ArgumentParser) -> str:
    """Validate the source given."""
    if not (options.source_pos or options.source):
        engine.error(
            'source path must be given - either as first positional argument or as value to the --source option'
        )

    if options.source_pos and options.source:
        engine.error(
            'source path given both as first positional argument and as value to the --source option - pick one'
        )

    if not options.source:
        options.source = options.source_pos

    if not pathlib.Path(options.source).is_file():
        engine.error(f'requested source ({options.source}) is not a file')

    source_suffix = pathlib.Path(options.source).suffix.lower()
    if source_suffix not in KNOWN_SUFFIXES:
        engine.error(
            f'requested source suffix ({source_suffix}) is not in known suffixes ({", ".join(sorted(KNOWN_SUFFIXES))})'
        )

    return source_suffix


def validate_target(options: argparse.Namespace, engine: argparse.ArgumentParser) -> str:
    """Validate the target given."""
    if not (options.target_pos or options.target):
        engine.error(
            'target path must be given - either as second positional argument or as value to the --target option'
        )

    if options.target_pos and options.target:
        engine.error(
            'target path given both as second positional argument and as value to the --target option - pick one'
        )

    if not options.target:
        options.target = options.target_pos

    target_suffix = pathlib.Path(options.target).suffix.lower()
    if target_suffix not in KNOWN_SUFFIXES:
        engine.error(
            f'requested target suffix ({target_suffix}) is not in known suffixes ({", ".join(sorted(KNOWN_SUFFIXES))})'
        )

    if pathlib.Path(options.target) == pathlib.Path(options.source):
        engine.error(
            'target path is identical with source path - cowardly giving up on the attempted in-place inversion'
        )

    return target_suffix


def parse_request(argv: list[str]) -> Union[int, argparse.Namespace]:
    """Parse the request vector into validated options."""
    parser = parser_configured()
    if not argv:
        parser.print_help()
        return 0

    options = parser.parse_args(argv)

    if options.version:
        print(VERSION)
        return 0

    # Ensure consistent request for verbosity
    if options.debug and options.quiet:
        parser.error('Cannot be quiet and debug - pick one')

    # Ensure hierarchical setting of debug: command line overwrites environment variable value
    if options.debug is None:
        options.debug = DEBUG

    # Quiet overwrites any debug requests
    if options.quiet:
        options.debug = False

    # Validate and derive the DSL values for source and target formats
    options.source_format = JSON_FORMAT if validate_source(options, engine=parser) in JSON_SUFFIXES else YAML_FORMAT
    options.target_format = JSON_FORMAT if validate_target(options, engine=parser) in JSON_SUFFIXES else YAML_FORMAT

    # Derive well-defined values for auto-serial mode related variables
    options.auto_serial_start = 1 if options.auto_serial else 0
    options.auto_serial_step = 1

    # Preview-only requested?
    if options.preview:
        print('Preview of inversion request (dry-run):')
        request = vars(options)
        for opt, val in request.items():
            print(f'  {opt}: {val}')
        return 0

    return options


def main(argv: Union[list[str], None] = None) -> int:
    """Delegate processing to functional module."""
    argv = sys.argv[1:] if argv is None else argv
    options = parse_request(argv)

    if isinstance(options, int):
        return 0
    return invert(request=options)
